/******************************************************************
Created with PROGRAMINO IDE for Arduino
Project     : ESP32 WiFi Camera
Board       : MAKERFACTORY EBB ESP32
Author      : MF

Description : Camera Webserver

Switches:
Camera      SW403 ON
Camera      SW402 ON
Camera      SW400 ON
I2C PULL-UP SW307 ON
I2C PULL-UP SW308 ON

Original from:
https://bitluni.net

ToDo:
- insert ssid and password
Changes:
- add defines for MAKERFACTORY EBB ESP32
- file: Fifo -> static unsigned char inline readByte() add 10µs delay for valid data
******************************************************************/

#include <WiFi.h>
#include <WiFiClient.h>

//#define USE_TFT_ESP32
//#define USE_TFT_NodeMcu
#define USE_EBB_ESP32

//Wifi credentials
#define ssid        "xxxxxxxxxxxxxxxx"
#define password    "xxxxxxxxxxxxxxxx"

#include "I2C.h"
#include "FifoCamera.h"
#include "BMP.h"

#include "Pins.h"

I2C<SIOD, SIOC> i2c;
FifoCamera<I2C<SIOD, SIOC>, RRST, WRST, RCK, WR, D0, D1, D2, D3, D4, D5, D6, D7> camera(i2c);

#ifdef USE_TFT_ESP32
    #include <SPI.h>
    #include <Adafruit_GFX.h>    
    #include <Adafruit_ILI9341.h> 
    Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS,  TFT_DC, TFT_RST);
    #include "Display.h"
#endif

#ifdef USE_TFT_NodeMcu
    #include <SPI.h>
    #include <Adafruit_GFX.h>    
    #include <Adafruit_ILI9341.h> 
    Adafruit_ILI9341 tft = Adafruit_ILI9341(TFT_CS,  TFT_DC, TFT_RST);
    #include "Display.h"
#endif

#ifdef USE_EBB_ESP32
    // nothing...
#endif

const int XRES = 160;
const int YRES = 120;

const int BYTES_PER_PIXEL = 2;
unsigned char bmpHeader[BMP::headerSize];
unsigned char frame[XRES  * YRES * BYTES_PER_PIXEL];

//---------------------Webserver----------------------------

WiFiServer server(80);

void Webserver_Start()
{
    server.begin();     // Start TCP/IP-Server on ESP32
}

// Put ESP32 in both modes in parallel: Soft Access Point and station in router network (LAN)
void WiFi_SetBothModesNetworkStationAndAccessPoint()
{
    WiFi.mode(WIFI_AP_STA);
}

// Initialize Soft Access Point with ESP32
// ESP32 establishes its own WiFi network, one can choose the SSID
void WiFi_AccessPointStart(char* AccessPointNetworkSSID)
{ 
    WiFi.softAP(AccessPointNetworkSSID);
    
    // printout the ESP32 IP-Address in own network, per default it is "192.168.4.1".
    Serial.print("AccessPointIP: ");
    Serial.println(WiFi.softAPIP());
    
    return;
}

// Connect to router network and return 1 (success) or -1 (no success)
int WiFi_RouterNetworkConnect(char* txtSSID, char* txtPassword)
{
    int success = 1;
    
    // connect to WiFi network
    // see https://www.arduino.cc/en/Reference/WiFiBegin
    
    WiFi.begin(txtSSID, txtPassword);
    delay(1000);
    
    // not connected
    if (WiFi.status() != WL_CONNECTED)
    {
        success = -1;
    }
    else
    {
        // print out local address of ESP32 in Router network (LAN)
        Serial.print("Local IP-Adress: ");
        Serial.println(WiFi.localIP());
    }
    
    return success;
}

void setup() 
{
    Serial.begin(115200);
    WiFi_SetBothModesNetworkStationAndAccessPoint();
    
    #if defined(USE_TFT_ESP32) || defined(USE_TFT_NodeMcu)
        tft.begin();
        tft.setRotation(1);
        tft.fillScreen(ILI9341_BLACK );
    #endif
    
    // Start access point with SSID "ESP32_MyJourneyIoT"
    WiFi_AccessPointStart("ESP32_MyJourneyIoT");  
    
    int success = 0;
    
    for(int i = 0; i < 3; i++)
    {
        success = WiFi_RouterNetworkConnect(ssid, password);
        if (success == 1) 
        {
            Serial.println("Connection to the given network established.");
            break;
        }
    }
    
    if (success == -1)
    {
        Serial.println("Connection to the given network could not be established.");
    }
    
    // start the webserver to listen for request of clients (in LAN or own ESP32 network)
    Webserver_Start();
    Serial.println("Server start");
    
    i2c.init();
    camera.init();
    
    BMP::construct16BitHeader(bmpHeader, XRES, YRES);
    camera.QQVGARGB565();
    //camera.QQQVGARGB565();
    //camera.QQVGAYUV();
    
    pinMode(VSYNC, INPUT);
    
    #ifdef USE_TFT_ESP32
        Serial.println("Reset display.");
        tft.fillScreen(0);
        testTFT(XRES, YRES);
    #endif
    
    #ifdef USE_TFT_NodeMcu
        Serial.println("Reset display.");
        tft.setRotation(3);
        tft.fillScreen(0);
    #endif
    
    #ifdef USE_EBB_ESP32
        // nothing...
    #endif
    
    
}

void serve()
{
    WiFiClient client = server.available();
    if (client) 
    {
        //Serial.println("New Client.");
        String currentLine = "";
        while (client.connected()) 
        {
            if (client.available()) 
            {
                char c = client.read();
                //Serial.write(c);
                if (c == '\n') 
                {
                    if (currentLine.length() == 0) 
                    {
                        client.println("HTTP/1.1 200 OK");
                        client.println("Content-type:text/html");
                        client.println();
                        client.print(
                        "<style>body{margin: 0}\nimg{height: 100%; width: auto}</style>"
                        "<img id='a' src='/camera' onload='this.style.display=\"initial\"; var b = document.getElementById(\"b\"); b.style.display=\"none\"; b.src=\"camera?\"+Date.now(); '>"
                        "<img id='b' style='display: none' src='/camera' onload='this.style.display=\"initial\"; var a = document.getElementById(\"a\"); a.style.display=\"none\"; a.src=\"camera?\"+Date.now(); '>");
                        client.println();
                        break;
                    } 
                    else 
                    {
                        currentLine = "";
                    }
                } 
                else if (c != '\r') 
                {
                    currentLine += c;
                }
                
                if(currentLine.endsWith("GET /camera"))
                {
                    client.println("HTTP/1.1 200 OK");
                    client.println("Content-type:image/bmp");
                    client.println();
                    
                    // Write the header to the client
                    client.write(bmpHeader, BMP::headerSize);
                    
                    // Write the picture to the client
                    client.write(frame,  XRES * YRES * BYTES_PER_PIXEL);
                    
                }
            }
        }
        // close the connection:
        client.stop();
        Serial.println("Client Disconnected.");
    }  
}

void readFrame()
{
    while(!digitalRead(VSYNC));
    while(digitalRead(VSYNC));
    camera.stopCapture();
    camera.prepareCapture();
    while(!digitalRead(VSYNC));
    camera.startCapture();
    camera.readFrame(frame, XRES, YRES, BYTES_PER_PIXEL);
}

void loop() 
{
    readFrame();
    
    #ifdef USE_TFT_ESP32
        displayRGB565(frame, XRES, YRES);
    #endif
    
    #ifdef USE_TFT_NodeMcu
        displayRGB565(frame, XRES, YRES);
    #endif
    
    #ifdef USE_EBB_ESP32
        // nothing...
    #endif
    
    serve();
}
